!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT, Inria
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

! subroutine zqrm_sytrf_old(uplo, m, k, a, lda, info)

!   use qrm_parameters_mod
!   implicit none
  
!   integer      :: lda, m, k, nb, i
!   character    :: uplo
!   complex(r64) :: a(lda,*)
!   integer      :: info

!   ! For now, we fix this block size statically in the source code. The future
!   ! shall tell us if it is pertinent to let the user change it through an
!   ! option.
!   nb = 32

!   if(k.lt.nb) nb = k
  
!   info = 0

!   if(uplo.eq.qrm_upper) then
!      do i=1, k, nb
!         call zsytrf_old
!         if(info.ne.0) return
!         if(m.ge.i+nb) then
!            call ztrsm(qrm_left, uplo,    &
!                 qrm_transp, qrm_no_unit, &
!                 nb, m - (i + nb - 1),    &
!                 qrm_zone,                &
!                 a(i,i), lda,             &
!                 a(i,i+nb), lda)
           
!            call zsyrk(uplo, qrm_transp, &
!                 m - (i + nb - 1), nb,   &
!                 qrm_zmone,              &
!                 a(i,i+nb), lda,         &
!                 qrm_zone,               &
!                 a(i+nb,i+nb), lda)
!         end if
!      end do
!   else
!      write(*,*)'qrm_sytrf with uplo=l not yet implemented'
!   end if

! contains
  
!   subroutine zsytrf_old
    
!     integer      :: j, ubound
!     complex(r64) :: alpha
    
!     ubound = min(i + nb - 1, k)

!     do j = i, ubound - 1
!        a(j,j) = sqrt(a(j,j))
!        if(a(j,j).eq.0) then
!           info = j
!           return
!        end if
!        alpha = qrm_zone / a(j,j)
!        call zscal(ubound - j, &
!             alpha, &
!             a(j,j+1), &
!             lda)
!        call zsyrk(uplo, qrm_transp, &
!             ubound - j, 1, &
!             qrm_zmone, &
!             a(j,j+1), lda, &
!             qrm_zone, &
!             a(j+1,j+1), lda)
!     end do

!     ! Factorize the last diagonal element.
!     a(j,j) = sqrt(a(j,j))
!     if(a(j,j).eq.0) then
!        info = j
!        return
!     end if

!   end subroutine zsytrf_old

! end subroutine zqrm_sytrf_old





subroutine zqrm_sytrf(uplo, m, k, a, lda, info)

  use qrm_parameters_mod
  implicit none
  
  integer      :: lda, m, k
  character    :: uplo
  complex(r64) :: a(lda,*)
  integer      :: info

  integer      :: i, mm, nb

  ! For now, we fix this block size statically in the source code. The future
  ! shall tell us if it is pertinent to let the user change it through an
  ! option.
  nb = 32

  if(k.lt.nb) nb = k
  
  info = 0

  if(uplo.eq.qrm_upper) then
     do i=1, k, nb
        mm = min(nb, k-i+1)
        call zsytrf_nopiv(mm, a(i,i), lda, info)
        if(info.ne.0) return
        if(m.ge.i+mm) then
           call ztrsm(qrm_left, uplo,    &
                qrm_transp, qrm_no_unit, &
                mm, m - (i + mm - 1),    &
                qrm_zone,                &
                a(i,i), lda,             &
                a(i,i+mm), lda)
           
           call zsyrk(uplo, qrm_transp, &
                m - (i + mm - 1), mm,   &
                qrm_zmone,              &
                a(i,i+mm), lda,         &
                qrm_zone,               &
                a(i+mm,i+mm), lda)
        end if
     end do
  else
     write(*,*)'qrm_sytrf with uplo=l not yet implemented'
  end if

end subroutine zqrm_sytrf
  
subroutine zsytrf_nopiv(m, a, lda, info)

  use qrm_parameters_mod
  implicit none
  
  integer      :: lda, m
  complex(r64) :: a(lda,*)
  integer      :: info

  integer      :: i
  complex(r64) :: alpha

  do i = 1, m
     if(a(i,i).eq.0) then
        info = i
        return
     end if
     a(i,i) = sqrt(a(i,i))
     if(i.lt.m) then
        alpha = qrm_zone / a(i,i)
        call zscal(m - i, &
             alpha,       &
             a(i,i+1), lda)
        call zsyrk('u', qrm_transp,  &
             m-i, 1,                 &
             qrm_zmone,              &
             a(i,i+1), lda,          &
             qrm_zone,               &
             a(i+1,i+1), lda)
     end if
  end do
  
  return
  
end subroutine zsytrf_nopiv



