!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a fill of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!! 
!! This file contains a routine for filling a block
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine cqrm_block_addi_task(qrm_dscr, aij, i, j, m, n, v, prio)
#if defined(have_starpu)
  use cqrm_starpu_mod
  use fstarpu_mod
#endif
  use qrm_dscr_mod
  use cqrm_dsmat_mod
  use qrm_string_mod
  use iso_c_binding
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(qrm_dscr_type), target    :: qrm_dscr
  type(cqrm_block_type)          :: aij
  integer, target                :: i, j, m, n
  complex(r32), target           :: v
  integer, optional              :: prio
#if defined (have_starpu)

  type(c_ptr), target            :: dscrpt
  type(c_ptr)                    :: sizeof
#endif
  integer(kind=8)                :: ts, te
  integer, target                :: iprio
  
  if(qrm_dscr%info.ne.0) return
  
#if defined (have_starpu)
  if(qrm_dscr%seq.eq.0) then
 
#if defined(dprec)
     sizeof = FSTARPU_SZ_REAL8
#elif defined(sprec)
     sizeof = FSTARPU_SZ_REAL4
#elif defined(zprec)
     sizeof = FSTARPU_SZ_COMPLEX8
#elif defined(cprec)
     sizeof = FSTARPU_SZ_COMPLEX4
#endif
       
     dscrpt = c_loc(qrm_dscr)
     iprio = qrm_starpu_get_prio(addi_, qrm_dscr%pr, prio)

     call fstarpu_insert_task    ((/                                &
          cqrm_block_addi_cl,                                       &
          FSTARPU_VALUE, c_loc(dscrpt),       FSTARPU_SZ_C_PTR,     &
          FSTARPU_VALUE, c_loc(i),            FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(j),            FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(m),            FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(n),            FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(v),            sizeof,               &
          FSTARPU_W, aij%hdl,                                       &
          FSTARPU_PRIORITY, c_loc(iprio),                           &
          FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                   &
          C_NULL_PTR /))

  else
#endif
     __QRM_PROF_START(ts)
     call cqrm_addi(aij%c(1,1), size(aij%c,1), i, j, m, n, v)
     
     __QRM_PROF_STOP(qrm_prof_addi_task_,ts,te)
#if defined (have_starpu)
  end if
#endif

  return
end subroutine cqrm_block_addi_task


#if defined (have_starpu)
recursive subroutine cqrm_starpu_block_addi_cpu_func(buffers, cl_arg) bind(C)
  use fstarpu_mod
  use cqrm_dsmat_mod
  use qrm_dscr_mod
#if defined (perfprof)
    use qrm_perfprof_mod
#endif

  implicit none

  type(c_ptr), value, intent(in) :: cl_arg
  type(c_ptr), value             :: buffers

  type(c_ptr), target            :: dscr_c
  type(qrm_dscr_type), pointer   :: qrm_dscr
  character, target              :: init
  complex(r32), target           :: v
  complex(r32), pointer          :: a(:,:)
  integer, target                :: i, j, m, n, lda
  integer(kind=8)                :: ts, te
  integer, target                :: worker_id
  
  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), &       
       c_loc(i), c_loc(j),                         &
       c_loc(m), c_loc(n),                         &
       c_loc(v) /))
  
  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)

  if(qrm_dscr%info.ne.0) return

  lda = fstarpu_matrix_get_ld(buffers, 0)

  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), a, shape=(/lda,n/))

  __QRM_PROF_START(ts)
  call cqrm_addi(a(1,1), lda, i, j, m, n, v)
  __QRM_PROF_STOP(qrm_prof_addi_task_,ts,te)
  return

end subroutine cqrm_starpu_block_addi_cpu_func
#endif
