!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains the module which holds the data types used during the factorization.
!!
!! ##############################################################################################


#include "qrm_common.h"

!> @brief This module contains the definition of a dense matrix
!!
module zqrm_dsmat_task_mod

  use iso_c_binding

  interface qrm_gemm_task
     subroutine zqrm_gemm_task(qrm_dscr, transa, transb, m, n, k, alpha, a, b, beta, c, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       type(zqrm_block_type)       :: a, b, c
       integer                     :: m, n, k
       complex(r64), target        :: alpha, beta
       character, target           :: transa, transb
       type(c_ptr)                 :: sizeof
       integer, optional           :: prio
     end subroutine zqrm_gemm_task
  end interface qrm_gemm_task

  interface qrm_higeqrt_task
     subroutine zqrm_higeqrt_task(qrm_dscr, m, n, nb, ib, k, a, t, work, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       type(zqrm_block_type)       :: a, t
       integer                     :: m, n, ib, k, nb
       type(zqrm_ws_type)          :: work
       integer, optional           :: prio
     end subroutine zqrm_higeqrt_task
  end interface qrm_higeqrt_task

  interface qrm_higmeqrt_task
     subroutine zqrm_higemqrt_task(qrm_dscr, trans, m, n, k, nb, ib, i, j, a, t, c, work, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       character                   :: trans
       type(zqrm_block_type)       :: a, t, c
       integer                     :: m, n, k, i, j, nb
       type(zqrm_ws_type)          :: work
       integer                     :: ib
       integer, optional           :: prio
     end subroutine zqrm_higemqrt_task
  end interface qrm_higmeqrt_task

  interface qrm_hitpqrt_task
     subroutine zqrm_hitpqrt_task(qrm_dscr, m, n, l, nb, ib, i, a1, a2, t, work, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type)         :: qrm_dscr
       type(zqrm_block_type)       :: a1, a2, t
       integer                     :: m, n, l, nb, ib, i
       type(zqrm_ws_type)          :: work
       integer, optional           :: prio
     end subroutine zqrm_hitpqrt_task
  end interface qrm_hitpqrt_task

  interface qrm_hitpmqrt_task
     subroutine zqrm_hitpmqrt_task(qrm_dscr, trans, m, n, k, l, nb, ib, i, j, v, t, c1, c2, work, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       character, target           :: trans
       type(zqrm_block_type)       :: v, c1, c2, t
       integer                     :: m, n, k, l, i, j, nb, ib
       type(zqrm_ws_type)          :: work
       integer, optional           :: prio
     end subroutine zqrm_hitpmqrt_task
  end interface qrm_hitpmqrt_task

  interface qrm_trsm_task
     subroutine zqrm_trsm_task(qrm_dscr, side, uplo, transa, diag, m, n, k, alpha, a, b, prio)
       use qrm_parameters_mod
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       character, target           :: side, uplo, transa, diag
       integer                     :: m, n, k
       complex(r64)                :: alpha
       type(zqrm_block_type)       :: a, b
       integer, optional           :: prio
     end subroutine zqrm_trsm_task
  end interface qrm_trsm_task

  interface qrm_trmm_task
     subroutine zqrm_trmm_task(qrm_dscr, side, uplo, transa, diag, m, n, k, alpha, a, b, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       character, target           :: side, uplo, transa, diag
       integer                     :: m, n, k
       complex(r64), target        :: alpha
       type(zqrm_block_type)       :: a, b
       integer, optional           :: prio
     end subroutine zqrm_trmm_task
  end interface qrm_trmm_task

  interface qrm_block_copy_task
     subroutine zqrm_block_copy_task(qrm_dscr, transp, a, b, ia, ja, ib, jb, m, n, l, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type), target    :: qrm_dscr
       character                      :: transp
       type(zqrm_block_type), target  :: a
       type(zqrm_block_type), target  :: b
       integer                        :: ia, ja, ib, jb
       integer                        :: m, n, l
       integer, optional              :: prio
     end subroutine zqrm_block_copy_task
  end interface qrm_block_copy_task

  interface qrm_block_axpy_task
     subroutine zqrm_block_axpy_task(qrm_dscr, alpha, a, b, ia, ja, ib, jb, m, n, l, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type), target    :: qrm_dscr
       complex(r64)                   :: alpha
       type(zqrm_block_type), target  :: a
       type(zqrm_block_type), target  :: b
       integer                        :: ia, ja, ib, jb
       integer                        :: m, n, l
       integer, optional              :: prio
     end subroutine zqrm_block_axpy_task
  end interface qrm_block_axpy_task

  interface qrm_block_extadd_task
     subroutine zqrm_block_extadd_task(qrm_dscr, a, b, br, bc, i, j, m, n, l, inout, op, rowmap, colmap &
#if defined(have_starpu)
          ,hdl &
#endif
          , prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type), target    :: qrm_dscr
       type(zqrm_dsmat_type), target  :: a
       type(zqrm_dsmat_type), target  :: b
       integer                        :: br, bc
       integer                        :: i, j, m, n, l
       character                      :: inout, op
       integer                        :: rowmap(:,:)
       integer, optional              :: colmap(:,:)
#if defined(have_starpu)
       type(c_ptr), optional          :: hdl
#endif
       integer, optional              :: prio
     end subroutine zqrm_block_extadd_task
  end interface qrm_block_extadd_task

!   interface qrm_block_extadd_byblock_task
!      subroutine zqrm_block_extadd_byblock_task(qrm_dscr, a, b, br, bc, i, j, m, n, l, inout, op, rowmap, colmap &
! #if defined(have_starpu)
!           ,hdl &
! #endif
!           , prio)
!        use qrm_dscr_mod
!        use zqrm_dsmat_mod
!        type(qrm_dscr_type), target    :: qrm_dscr
!        type(zqrm_dsmat_type), target  :: a
!        type(zqrm_dsmat_type), target  :: b
!        integer                        :: br, bc
!        integer                        :: i, j, m, n, l
!        character                      :: inout, op
!        integer                        :: rowmap(:,:)
!        integer, optional              :: colmap(:,:)
! #if defined(have_starpu)
!        type(c_ptr), optional          :: hdl
! #endif
!        integer, optional              :: prio
!      end subroutine zqrm_block_extadd_byblock_task
!   end interface qrm_block_extadd_byblock_task

  interface qrm_block_nrm_task
     subroutine zqrm_block_nrm_task(qrm_dscr, a, m, n, ssq, prio)
       use zqrm_dsmat_mod
       use qrm_dscr_mod
       type(qrm_dscr_type), target   :: qrm_dscr
       type(zqrm_block_type), target :: a
       integer, target               :: m, n
       type(zqrm_ws_type)            :: ssq
       integer, optional             :: prio
     end subroutine zqrm_block_nrm_task
  end interface qrm_block_nrm_task

  interface qrm_block_trdcn_task
     subroutine zqrm_block_trdcn_task(qrm_dscr, a, n, d, eps)
       use zqrm_dsmat_mod
       use qrm_dscr_mod
       type(qrm_dscr_type), target   :: qrm_dscr
       type(zqrm_block_type), target :: a
       integer, target               :: n, d
       real(r64)                     :: eps
     end subroutine zqrm_block_trdcn_task
  end interface qrm_block_trdcn_task

  interface qrm_potrf_task
     subroutine zqrm_potrf_task(qrm_dscr, uplo, m, k, a, work, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       character, target           :: uplo
       integer, target             :: m, k
       type(zqrm_ws_type)          :: work
       type(zqrm_block_type)       :: a
       integer, optional           :: prio
     end subroutine zqrm_potrf_task
  end interface qrm_potrf_task

  interface qrm_sytrf_task
     subroutine zqrm_sytrf_task(qrm_dscr, uplo, m, k, a, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       character, target           :: uplo
       integer, target             :: m, k
       type(zqrm_block_type)       :: a
       integer, optional           :: prio
     end subroutine zqrm_sytrf_task
  end interface qrm_sytrf_task

  interface qrm_herk_task
     subroutine zqrm_herk_task(qrm_dscr, uplo, transa, n, k, alpha, a, beta, c, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type)         :: qrm_dscr
       character                   :: uplo, transa
       integer                     :: n, k
       complex(r64)                :: alpha, beta
       type(zqrm_block_type)       :: a, c
       integer, optional           :: prio
     end subroutine zqrm_herk_task
  end interface qrm_herk_task

  interface qrm_syrk_task
     subroutine zqrm_syrk_task(qrm_dscr, uplo, transa, n, k, alpha, a, beta, c, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type)         :: qrm_dscr
       character                   :: uplo, transa
       integer                     :: n, k
       complex(r64)                :: alpha, beta
       type(zqrm_block_type)       :: a, c
       integer, optional           :: prio
     end subroutine zqrm_syrk_task
  end interface qrm_syrk_task



  interface qrm_block_fill_task
     subroutine zqrm_block_fill_task(qrm_dscr, aij, i, j, m, n, l, init, ivalue, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type)    :: qrm_dscr
       type(zqrm_block_type)  :: aij
       integer                :: i, j, m, n, l
       character              :: init
       complex(r64)           :: ivalue
       integer, optional      :: prio
     end subroutine zqrm_block_fill_task
  end interface qrm_block_fill_task


  interface qrm_block_zero_task
     subroutine zqrm_block_zero_task(qrm_dscr, aij, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type)    :: qrm_dscr
       type(zqrm_block_type)  :: aij
       integer, optional      :: prio
     end subroutine zqrm_block_zero_task
  end interface qrm_block_zero_task



  interface qrm_block_addi_task
     subroutine zqrm_block_addi_task(qrm_dscr, aij, i, j, m, n, v, prio)
       use qrm_dscr_mod
       use zqrm_dsmat_mod
       type(qrm_dscr_type), target    :: qrm_dscr
       type(zqrm_block_type)          :: aij
       integer, target                :: i, j, m, n
       complex(r64), target           :: v
       integer, optional              :: prio
     end subroutine zqrm_block_addi_task
  end interface qrm_block_addi_task
  
  
end module zqrm_dsmat_task_mod
