!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine zqrm_hitpmqrt_task(qrm_dscr, trans, m, n, k, l, nb, ib, i, j, v, t, c1, c2, work, prio)

#if defined(have_starpu)
  use zqrm_starpu_mod
  use qrm_starpu_common_mod
#endif
  use zqrm_dsmat_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(qrm_dscr_type), target   :: qrm_dscr
  character, target             :: trans
  type(zqrm_block_type), target :: v, c1, c2, t
  integer, target               :: m, n, k, l, nb, ib, i, j
  type(zqrm_ws_type)            :: work
  integer, optional             :: prio

#if defined(have_starpu)
  type(c_ptr)                   :: hdlv, hdlc1, hdlc2, hdlt
  type(c_ptr), target           :: dscrpt, strpt
  real(c_double),target         :: flops
#endif
  integer, target               :: ii, e, iprio
  integer                       :: jj, ldc1, ldc2, ldt, ldv, nc1, nv, info
  integer(kind=8)               :: tb, te

  if(qrm_dscr%info.ne.0) return

  if(v%partitioned) then
     ii    = (i-1)*nb+1
  else
     ii    = 1
  end if

#if defined (have_starpu)

  if(qrm_dscr%seq.eq.0) then

     ! submit a geqrt task
     if(v%partitioned) then
        hdlv  = v%shdls(i)
        hdlt  = t%shdls(i)
     else
        hdlv  = v%hdl
        hdlt  = t%hdl
     end if

     if(c1%partitioned) then
        hdlc1 = c1%shdls(j)
        hdlc2 = c2%shdls(j)
        iprio = qrm_starpu_get_prio(tpmqrt_i_, qrm_dscr%pr, prio)
     else
        hdlc1 = c1%hdl
        hdlc2 = c2%hdl
        iprio = qrm_starpu_get_prio(tpmqrt_o_, qrm_dscr%pr, prio)
     end if

     dscrpt = c_loc(qrm_dscr)
     if(qrm_allocated(v%stair)) then
        strpt = c_loc(v%stair(ii))
     else
        strpt = c_null_ptr
     end if

     flops = 4.d0*real(m)*real(n)*real(k) - 2.d0*real(n)*real(l)*real(l)
     ! write(*,*)'tpmqrt ',m,n,k,l,flops
     
     call fstarpu_insert_task((/zqrm_tpmqrt_cl,               &
          FSTARPU_VALUE, c_loc(dscrpt), FSTARPU_SZ_C_PTR,     &
          FSTARPU_VALUE, c_loc(m),      FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(n),      FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(k),      FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(l),      FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(ib),     FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(ii),     FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(trans),  FSTARPU_SZ_CHARACTER, &
          FSTARPU_VALUE, c_loc(strpt),  FSTARPU_SZ_C_PTR,     &
          FSTARPU_R, hdlv,                                    &
          FSTARPU_RW, hdlc1,                                  &
          FSTARPU_RW, hdlc2,                                  &
          FSTARPU_R, hdlt,                                    &
          FSTARPU_SCRATCH, work%hdl,                          &
          FSTARPU_PRIORITY, c_loc(iprio),                     &
          ! FSTARPU_FLOPS, c_loc(flops),                        &
          FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),             &
          C_NULL_PTR /))
     ! call fstarpu_task_wait_for_all()
  else
#endif

     ! make a simple call to lapack tpmqrt
     ldc1 = size(c1%c,1)
     ldc2 = size(c2%c,1)
     ldv  = size(v%c,1)
     ldt  = size(t%c,1)
     jj   = (j-1)*nb+1

     __QRM_PROF_START(tb)
     if(qrm_allocated(v%stair)) then
        call zqrm_tpmqrt(qrm_left, trans,      &
                         m, n, k, l, ib,       &
                         v%stair(ii),          &
                         v%c(1,ii),   ldv,     &
                         t%c(1,ii),   ldt,     &
                         c1%c(ii,jj), ldc1,    &
                         c2%c(1,jj),  ldc2,    &
                         work%c(1,1), info)
     else
        call zqrm_tpmqrt(qrm_left, trans,      &
                         m, n, k, l, ib,       &
                         -1,                   &
                         v%c(1,ii),   ldv,     &
                         t%c(1,ii),   ldt,     &
                         c1%c(ii,jj), ldc1,    &
                         c2%c(1,jj),  ldc2,    &
                         work%c(1,1), info)
     end if
     __QRM_PROF_STOP(qrm_prof_tpmqrt_task_,tb,te)

#if defined (have_starpu)
  end if
#endif

  return
end subroutine zqrm_hitpmqrt_task



#if defined (have_starpu)
recursive subroutine zqrm_starpu_tpmqrt_cpu_func(buffers, cl_arg) bind(C)
  use qrm_parameters_mod
  use qrm_dscr_mod
  use qrm_starpu_common_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value, intent(in) :: cl_arg
  type(c_ptr), value             :: buffers

  integer(c_int), target         :: m, n, k, l, ib, ii
  character, target              :: trans
  type(c_ptr), target            :: dscr_c, stair_c
  integer                        :: mv, nv, ldv, mc1, nc1, ldc1, mc2, nc2, ldc2, nt, ldt, nw, ldw
  integer                        :: info, jj
  complex(r64), pointer          :: v(:,:), c1(:,:), c2(:,:), t(:,:), w(:,:)
  integer, pointer               :: stair(:)
  type(qrm_dscr_type), pointer   :: qrm_dscr
  integer(kind=8)                :: tb, te


  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), &
                                    c_loc(m),      &
                                    c_loc(n),      &
                                    c_loc(k),      &
                                    c_loc(l),      &
                                    c_loc(ib),     &
                                    c_loc(ii),     &
                                    c_loc(trans),  &
                                    c_loc(stair_c) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  nv    = fstarpu_matrix_get_ny(buffers, 0)
  ldv   = fstarpu_matrix_get_ld(buffers, 0)
  nc1   = fstarpu_matrix_get_ny(buffers, 1)
  ldc1  = fstarpu_matrix_get_ld(buffers, 1)
  nc2   = fstarpu_matrix_get_ny(buffers, 2)
  ldc2  = fstarpu_matrix_get_ld(buffers, 2)
  nt    = fstarpu_matrix_get_ny(buffers, 3)
  ldt   = fstarpu_matrix_get_ld(buffers, 3)
  nw    = fstarpu_matrix_get_ny(buffers, 4)
  ldw   = fstarpu_matrix_get_ld(buffers, 4)

  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), v,  shape=(/ldv,nv/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 1), c1, shape=(/ldc1,nc1/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 2), c2, shape=(/ldc2,nc2/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 3), t,  shape=(/ldt,nt/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 4), w,  shape=(/ldw,nw/))

  __QRM_PROF_START(tb)
  if(c_associated(stair_c)) then
     call c_f_pointer(stair_c, stair, shape=(/nv/))
     call zqrm_tpmqrt(qrm_left, trans, &
          m, n, k, l, ib,              &
          stair(1),                    &
          v(1,1),   ldv,               &
          t(1,1),   ldt,               &
          c1(ii,1), ldc1,              &
          c2(1,1),  ldc2,              &
          w(1,1), info)
  else
     call zqrm_tpmqrt(qrm_left, trans, &
          m, n, k, l, ib,              &
          -1,                          &
          v(1,1),   ldv,               &
          t(1,1),   ldt,               &
          c1(ii,1), ldc1,              &
          c2(1,1),  ldc2,              &
          w(1,1), info)
  end if
  __QRM_PROF_STOP(qrm_prof_tpmqrt_task_,tb,te)

  ! TODO: add check for return code in info

  return
end subroutine zqrm_starpu_tpmqrt_cpu_func

#if defined (have_cuda)
recursive subroutine zqrm_starpu_tpmqrt_cuda_func(buffers, cl_arg) bind(C)
  use qrm_parameters_mod
  use qrm_dscr_mod
  use qrm_starpu_common_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value, intent(in) :: cl_arg
  type(c_ptr), value             :: buffers

  integer(c_int), target         :: ii, m, n, k, l, ib
  character, target              :: trans
  type(c_ptr), target            :: dscr_c, stair_c, vp, tp, c1p, c2p, wp, handlep
  integer                        :: nv, ldv, nc1, ldc1, nc2, ldc2, nt, ldt, nw, ldw
  integer                        :: info, id, lw
  type(qrm_dscr_type), pointer   :: qrm_dscr
  integer(kind=8)                :: tb, te

  interface
     subroutine zqrm_cuda_tpmqrt(         &
          side, trans,                    &
          m, n, k, l, nb,                 &
          stair, i,                       &
          v, ldv,                         &
          t, ldt,                         &
          a, lda,                         &
          b, ldb,                         &
          w, ldw) bind(C)
       use iso_c_binding
       type(c_ptr), value       :: v, t, a, b, w, stair
       character(c_char), value :: side, trans
       integer, value           :: m, n, k, l, nb, i, ldt, lda, ldb, ldw, ldv
     end subroutine zqrm_cuda_tpmqrt
  end interface

  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), &
                                    c_loc(m),      &
                                    c_loc(n),      &
                                    c_loc(k),      &
                                    c_loc(l),      &
                                    c_loc(ib),     &
                                    c_loc(ii),     &
                                    c_loc(trans),  &
                                    c_loc(stair_c) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  nv    = fstarpu_matrix_get_ny(buffers, 0)
  ldv   = fstarpu_matrix_get_ld(buffers, 0)
  nc1   = fstarpu_matrix_get_ny(buffers, 1)
  ldc1  = fstarpu_matrix_get_ld(buffers, 1)
  nc2   = fstarpu_matrix_get_ny(buffers, 2)
  ldc2  = fstarpu_matrix_get_ld(buffers, 2)
  nt    = fstarpu_matrix_get_ny(buffers, 3)
  ldt   = fstarpu_matrix_get_ld(buffers, 3)
  nw    = fstarpu_matrix_get_ny(buffers, 4)
  ldw   = fstarpu_matrix_get_ld(buffers, 4)
  lw    = nw*ldw;

  vp    = fstarpu_matrix_get_ptr(buffers, 0)
  c1p   = fstarpu_matrix_get_ptr(buffers, 1)
  c2p   = fstarpu_matrix_get_ptr(buffers, 2)
  tp    = fstarpu_matrix_get_ptr(buffers, 3)
  wp    = fstarpu_matrix_get_ptr(buffers, 4)

  ! id      = fstarpu_worker_get_id();
  ! handlep = qrm_dscr%cublas_handles(id+1)

  ! TODO: add check for return code in info
  __QRM_PROF_START(tb)
  call zqrm_cuda_tpmqrt(         &
       qrm_left, trans,          &
       m, n, k, l, ib,           &
       stair_c, ii,              &
       vp, ldv,                  &
       tp, ldt,                  &
       c1p, ldc1,                &
       c2p, ldc2,                &
       wp, lw)
  __QRM_PROF_STOP(qrm_prof_tpmqrt_task_,tb,te)

  return
end subroutine zqrm_starpu_tpmqrt_cuda_func
#endif



#endif
