!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains the dqrm_factorization_mod with  the generic interfaces
!! for all the factorization related routines.
!!
!! ##############################################################################################


!> This module contains all the generic interfaces for the typed routines
!! in the factorization phase
module dqrm_factorization_mod

  !> Generic interface for the dqrm_factorization_init routine
  interface qrm_factorization_init
     subroutine dqrm_factorization_init(qrm_dscr, qrm_mat, qrm_spfct, transp)
       use dqrm_spmat_mod
       use dqrm_spfct_mod
       type(qrm_dscr_type), target              :: qrm_dscr
       type(dqrm_spmat_type), target            :: qrm_mat
       type(dqrm_spfct_type), target            :: qrm_spfct
       character, intent(in)                    :: transp
     end subroutine dqrm_factorization_init
  end interface

  interface qrm_factorization_init_core
     subroutine dqrm_factorization_init_core(qrm_mat, qrm_spfct, transp, info)
       use dqrm_spmat_mod
       use dqrm_spfct_mod
       type(dqrm_spmat_type)                    :: qrm_mat
       type(dqrm_spfct_type), target            :: qrm_spfct
       character, intent(in)                    :: transp
       integer, optional                        :: info
     end subroutine dqrm_factorization_init_core
  end interface

  !> Generic interface for the dqrm_factorize routine
  interface qrm_factorize
     subroutine dqrm_factorize(qrm_mat, qrm_spfct, transp, info)
       use dqrm_spmat_mod
       use dqrm_spfct_mod
       type(dqrm_spmat_type)                    :: qrm_mat
       type(dqrm_spfct_type)                    :: qrm_spfct
       character, optional, intent(in)          :: transp
       integer, optional                        :: info
     end subroutine dqrm_factorize
  end interface

  !> Generic interface for the dqrm_factorize_async routine
  interface qrm_factorize_async
     subroutine dqrm_factorize_async(qrm_dscr, qrm_mat, qrm_spfct, transp)
       use dqrm_spmat_mod
       use dqrm_spfct_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                      :: qrm_dscr
       type(dqrm_spmat_type), target            :: qrm_mat
       type(dqrm_spfct_type)                    :: qrm_spfct
       character, optional, intent(in)          :: transp
     end subroutine dqrm_factorize_async
  end interface

  !> Generic interface for the dqrm_factorization_core routine
  interface qrm_factorization_core
     subroutine dqrm_factorization_core(qrm_dscr, qrm_spfct)
       use dqrm_spfct_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                      :: qrm_dscr
       type(dqrm_spfct_type), target            :: qrm_spfct
     end subroutine dqrm_factorization_core
  end interface

  !> Generic interface for the dqrm_init_front routine
  interface qrm_init_front
     subroutine dqrm_init_front(qrm_spfct, front, info)
       use dqrm_fdata_mod
       use dqrm_spfct_mod
       type(dqrm_spfct_type), target            :: qrm_spfct
       type(dqrm_front_type)                    :: front
       integer, optional                        :: info
     end subroutine dqrm_init_front
  end interface qrm_init_front

  !> Generic interface for the dqrm_init_block routine
  interface qrm_init_block
     subroutine dqrm_init_block(qrm_spfct, fnum, br, bc, info)
       use dqrm_fdata_mod
       use dqrm_spfct_mod
       type(dqrm_spfct_type), target            :: qrm_spfct
       integer                                  :: fnum, br, bc
       integer, optional                        :: info
     end subroutine dqrm_init_block
  end interface qrm_init_block

  !> Generic interface for the dqrm_init_block routine
  interface qrm_clean_block
     subroutine dqrm_clean_block(qrm_spfct, fnum, br, bc, info)
       use dqrm_fdata_mod
       use dqrm_spfct_mod
       type(dqrm_spfct_type), target            :: qrm_spfct
       integer                                  :: fnum, br, bc
       integer, optional                        :: info
     end subroutine dqrm_clean_block
  end interface qrm_clean_block

  !> Generic interface for the dqrm_factorize_front routine
  interface qrm_activate_front
     subroutine dqrm_activate_front(qrm_spfct, front, work, info)
       use dqrm_fdata_mod
       use dqrm_spfct_mod
       type(dqrm_spfct_type), target            :: qrm_spfct
       type(dqrm_front_type)                    :: front
       integer, optional                        :: work(:)
       integer, optional                        :: info
     end subroutine dqrm_activate_front
  end interface qrm_activate_front


  !> Generic interface for the dqrm_do_subtree routine
  interface qrm_do_subtree
     subroutine dqrm_do_subtree(qrm_spfct, root, flops, info)
       use dqrm_spfct_mod
       use dqrm_fdata_mod
       use qrm_parameters_mod
       type(dqrm_spfct_type), target  :: qrm_spfct
       type(dqrm_front_type), target  :: root
       real(r32)                      :: flops
       integer, optional              :: info
     end subroutine dqrm_do_subtree
  end interface

  !> Generic interface for the dqrm_clean_front routine
  interface qrm_clean_front
     subroutine dqrm_clean_front(qrm_spfct, front, info)
       use dqrm_spfct_mod
       use dqrm_fdata_mod
       type(dqrm_spfct_type), target            :: qrm_spfct
       type(dqrm_front_type)                    :: front
       integer, optional                        :: info
     end subroutine dqrm_clean_front
  end interface


  !> Generic interface for the  dqrm_assemble_front routine
  interface qrm_assemble_front
     subroutine dqrm_assemble_front(qrm_dscr, qrm_spfct, front, child, prio)
       use dqrm_spfct_mod
       use dqrm_fdata_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)            :: qrm_dscr
       type(dqrm_spfct_type)          :: qrm_spfct
       type(dqrm_front_type)          :: front, child
       integer, optional              :: prio
     end subroutine dqrm_assemble_front
  end interface qrm_assemble_front
  
end module dqrm_factorization_mod
