!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains a module with generic interfaces for all the solve typed routines
!!
!! ##############################################################################################


!> @brief This module contains all the interfaces for the typed routines
!! in the solve phase
module sqrm_solve_mod

  interface qrm_apply
     procedure                              :: sqrm_spfct_unmqr2d, sqrm_spfct_unmqr1d
  end interface qrm_apply

  !> Generic interface for the sqrm_spfct_oldunmqr2d and sqrm_spfct_oldunmqr1d routines
  !!
  interface qrm_spfct_unmqr
     procedure                              :: sqrm_spfct_unmqr2d, sqrm_spfct_unmqr1d
  end interface qrm_spfct_unmqr

  interface qrm_apply_async
     procedure                              :: sqrm_spfct_unmqr_async
  end interface qrm_apply_async
  
  interface qrm_spfct_unmqr_async
     subroutine sqrm_spfct_unmqr_async(qrm_dscr, qrm_spfct, transp, sdata)
       use sqrm_spfct_mod
       use sqrm_sdata_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                  :: qrm_dscr
       type(sqrm_spfct_type), intent(in)    :: qrm_spfct
       type(sqrm_sdata_type)                :: sdata
       character(len=*), intent(in)         :: transp
     end subroutine sqrm_spfct_unmqr_async
  end interface qrm_spfct_unmqr_async

  interface sqrm_spfct_unmqr
     subroutine sqrm_spfct_unmqr2d(qrm_spfct, transp, b, x, info)
       use sqrm_spfct_mod
       type(sqrm_spfct_type), target       :: qrm_spfct
       character(len=*)                    :: transp
       real(r32), target                :: b(:,:)
       real(r32), optional, target      :: x(:,:)
       integer, optional                   :: info
     end subroutine sqrm_spfct_unmqr2d
     subroutine sqrm_spfct_unmqr1d(qrm_spfct, transp, b, x, info)
       use sqrm_spfct_mod
       type(sqrm_spfct_type), target  :: qrm_spfct
       character(len=*)               :: transp
       real(r32), target           :: b(:)
       real(r32), optional, target :: x(:)
       integer, optional              :: info
     end subroutine sqrm_spfct_unmqr1d
  end interface sqrm_spfct_unmqr


  !> Generic interface for the sqrm_spfct_trsm and sqrm_spfct_trsm1d  routines
  interface qrm_solve
     procedure                              :: sqrm_spfct_trsm1d, sqrm_spfct_trsm2d
  end interface qrm_solve

  interface qrm_solve_async
     procedure                              :: sqrm_spfct_trsm_async
  end interface qrm_solve_async

  !> Generic interface for the sqrm_spfct_trsm and sqrm_spfct_trsm1d  routines
  interface qrm_spfct_trsm
     procedure                              :: sqrm_spfct_trsm1d, sqrm_spfct_trsm2d
  end interface qrm_spfct_trsm

  interface qrm_spfct_trsm_async
     subroutine sqrm_spfct_trsm_async(qrm_dscr, qrm_spfct, transp, qrm_sdata)
       use sqrm_spfct_mod
       use sqrm_sdata_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                  :: qrm_dscr
       type(sqrm_spfct_type), intent(in)    :: qrm_spfct
       type(sqrm_sdata_type)                :: qrm_sdata
       character(len=*), intent(in)         :: transp
     end subroutine sqrm_spfct_trsm_async
  end interface qrm_spfct_trsm_async

  interface qrm_spfct_trsm_bu_async
     subroutine sqrm_spfct_trsm_bu_async(qrm_dscr, qrm_spfct, transp, sdata)
       use sqrm_spfct_mod
       use sqrm_sdata_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                        :: qrm_dscr
       type(sqrm_spfct_type), intent(in),  target :: qrm_spfct
       character(len=*), intent(in)               :: transp
       type(sqrm_sdata_type), target              :: sdata
     end subroutine sqrm_spfct_trsm_bu_async
  end interface qrm_spfct_trsm_bu_async

  interface qrm_spfct_trsm_td_async
     subroutine sqrm_spfct_trsm_td_async(qrm_dscr, qrm_spfct, transp, sdata)
       use sqrm_spfct_mod
       use sqrm_sdata_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                        :: qrm_dscr
       type(sqrm_spfct_type), intent(in),  target :: qrm_spfct
       character(len=*), intent(in)               :: transp
       type(sqrm_sdata_type), target              :: sdata
     end subroutine sqrm_spfct_trsm_td_async
  end interface qrm_spfct_trsm_td_async

  interface qrm_spfct_unmqr_bu_async
     subroutine sqrm_spfct_unmqr_bu_async(qrm_dscr, qrm_spfct, transp, sdata)
       use sqrm_spfct_mod
       use sqrm_sdata_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                        :: qrm_dscr
       type(sqrm_spfct_type), intent(in),  target :: qrm_spfct
       character(len=*), intent(in)               :: transp
       type(sqrm_sdata_type), target              :: sdata
     end subroutine sqrm_spfct_unmqr_bu_async
  end interface qrm_spfct_unmqr_bu_async

  interface qrm_spfct_unmqr_td_async
     subroutine sqrm_spfct_unmqr_td_async(qrm_dscr, qrm_spfct, transp, sdata)
       use sqrm_spfct_mod
       use sqrm_sdata_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                        :: qrm_dscr
       type(sqrm_spfct_type), intent(in),  target :: qrm_spfct
       character(len=*), intent(in)               :: transp
       type(sqrm_sdata_type), target              :: sdata
     end subroutine sqrm_spfct_unmqr_td_async
  end interface qrm_spfct_unmqr_td_async


  !> Generic interface for the sqrm_spfct_trsm2d and sqrm_spfct_trsm1d routines
  interface sqrm_spfct_trsm
     subroutine sqrm_spfct_trsm2d(qrm_spfct, transp, b, x, info)
       use sqrm_spfct_mod
       type(sqrm_spfct_type), target       :: qrm_spfct
       real(r32), target                :: b(:,:)
       real(r32), optional, target      :: x(:,:)
       character(len=*)                    :: transp
       integer, optional                   :: info
     end subroutine sqrm_spfct_trsm2d
     subroutine sqrm_spfct_trsm1d(qrm_spfct, transp, b, x, info)
       use sqrm_spfct_mod
       type(sqrm_spfct_type), target  :: qrm_spfct
       real(r32), target           :: b(:)
       real(r32), optional, target :: x(:)
       character(len=*)               :: transp
       integer, optional              :: info
     end subroutine sqrm_spfct_trsm1d
  end interface sqrm_spfct_trsm

  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Kernels
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  interface qrm_spfct_trsm_subtree
     subroutine sqrm_spfct_trsm_subtree(qrm_spfct, root, sdata, transp, info)
       use sqrm_spfct_mod
       use sqrm_sdata_mod
       type(sqrm_spfct_type), target  :: qrm_spfct
       type(sqrm_sdata_type), target  :: sdata
       integer                        :: root
       character(len=*), intent(in)   :: transp
       integer, optional              :: info
     end subroutine sqrm_spfct_trsm_subtree
  end interface qrm_spfct_trsm_subtree

  interface qrm_spfct_unmqr_subtree
     subroutine sqrm_spfct_unmqr_subtree(qrm_spfct, root, sdata, transp, work, info)
       use sqrm_spfct_mod
       use sqrm_sdata_mod
       type(sqrm_spfct_type), target  :: qrm_spfct
       type(sqrm_sdata_type), target  :: sdata
       integer                        :: root
       character(len=*), intent(in)   :: transp
       type(sqrm_ws_type)             :: work
       integer, optional              :: info
     end subroutine sqrm_spfct_unmqr_subtree
  end interface qrm_spfct_unmqr_subtree

  interface qrm_spfct_trsm_activate_front
     subroutine sqrm_spfct_trsm_activate_front(qrm_spfct, front, front_rhs, nb, info)
       use sqrm_dsmat_mod
       use sqrm_spfct_mod
       use sqrm_sdata_mod
       type(sqrm_spfct_type) :: qrm_spfct
       type(sqrm_front_type) :: front
       type(sqrm_rhs_type)   :: front_rhs
       integer               :: nb
       integer               :: info
     end subroutine sqrm_spfct_trsm_activate_front
  end interface qrm_spfct_trsm_activate_front

  interface qrm_spfct_unmqr_activate_front
     subroutine sqrm_spfct_unmqr_activate_front(qrm_spfct, front, front_rhs, nb, info)
       use sqrm_dsmat_mod
       use sqrm_spfct_mod
       use sqrm_sdata_mod
       type(sqrm_spfct_type) :: qrm_spfct
       type(sqrm_front_type) :: front
       type(sqrm_rhs_type)   :: front_rhs
       integer               :: nb
       integer               :: info
     end subroutine sqrm_spfct_unmqr_activate_front
  end interface qrm_spfct_unmqr_activate_front


  interface qrm_spfct_trsm_init_front
     subroutine sqrm_spfct_trsm_init_front(qrm_spfct, front, front_rhs, info)
       use qrm_dscr_mod
       use sqrm_spfct_mod
       use sqrm_fdata_mod
       use sqrm_sdata_mod
       type(sqrm_spfct_type) :: qrm_spfct
       type(sqrm_front_type) :: front
       type(sqrm_rhs_type)   :: front_rhs
       integer               :: info
     end subroutine sqrm_spfct_trsm_init_front
  end interface qrm_spfct_trsm_init_front

  interface qrm_spfct_unmqr_init_front
     subroutine sqrm_spfct_unmqr_init_front(qrm_spfct, front, front_rhs, info)
       use qrm_dscr_mod
       use sqrm_spfct_mod
       use sqrm_fdata_mod
       use sqrm_sdata_mod
       type(sqrm_spfct_type) :: qrm_spfct
       type(sqrm_front_type) :: front
       type(sqrm_rhs_type)   :: front_rhs
       integer               :: info
     end subroutine sqrm_spfct_unmqr_init_front
  end interface qrm_spfct_unmqr_init_front



  interface qrm_spfct_trsm_clean_front
     subroutine sqrm_spfct_trsm_clean_front(front, front_rhs, lhs, transp, info)
       use sqrm_spfct_mod
       use sqrm_fdata_mod
       use sqrm_sdata_mod
       type(sqrm_front_type) :: front
       type(sqrm_rhs_type)   :: front_rhs
       real(r32)          :: lhs(:,:)
       character(len=*)      :: transp
       integer               :: info
     end subroutine sqrm_spfct_trsm_clean_front
  end interface qrm_spfct_trsm_clean_front
  
  interface qrm_spfct_unmqr_clean_front
     subroutine sqrm_spfct_unmqr_clean_front(front, front_rhs, lhs, transp, info)
       use sqrm_spfct_mod
       use sqrm_fdata_mod
       use sqrm_sdata_mod
       type(sqrm_front_type) :: front
       type(sqrm_rhs_type)   :: front_rhs
       real(r32)          :: lhs(:,:)
       character(len=*)      :: transp
       integer               :: info
     end subroutine sqrm_spfct_unmqr_clean_front
  end interface qrm_spfct_unmqr_clean_front
  
  interface qrm_spfct_trsm_init_block
     subroutine sqrm_spfct_trsm_init_block(front, front_rhs, transp, br, bc, rhs, info)
       use sqrm_spfct_mod
       use sqrm_fdata_mod
       use sqrm_sdata_mod
       type(sqrm_front_type) :: front
       type(sqrm_rhs_type)   :: front_rhs
       character(len=*)      :: transp
       integer               :: br, bc
       real(r32)          :: rhs(:,:)
       integer               :: info
     end subroutine sqrm_spfct_trsm_init_block
  end interface qrm_spfct_trsm_init_block

  interface qrm_spfct_unmqr_init_block
     subroutine sqrm_spfct_unmqr_init_block(front, front_rhs, transp, br, bc, rhs, info)
       use sqrm_spfct_mod
       use sqrm_fdata_mod
       use sqrm_sdata_mod
       type(sqrm_front_type) :: front
       type(sqrm_rhs_type)   :: front_rhs
       character(len=*)      :: transp
       integer               :: br, bc
       real(r32)          :: rhs(:,:)
       integer               :: info
     end subroutine sqrm_spfct_unmqr_init_block
  end interface qrm_spfct_unmqr_init_block
  

  interface qrm_spfct_trsm_clean_block
     subroutine sqrm_spfct_trsm_clean_block(front, front_rhs, transp, br, bc, lhs, info)
       use sqrm_spfct_mod
       use sqrm_fdata_mod
       use sqrm_sdata_mod
       type(sqrm_front_type) :: front
       type(sqrm_rhs_type)   :: front_rhs
       character(len=*)      :: transp
       integer               :: br, bc
       real(r32)          :: lhs(:,:)
       integer               :: info
     end subroutine sqrm_spfct_trsm_clean_block
  end interface qrm_spfct_trsm_clean_block

  interface qrm_spfct_unmqr_clean_block
     subroutine sqrm_spfct_unmqr_clean_block(front, front_rhs, transp, br, bc, lhs, info)
       use sqrm_spfct_mod
       use sqrm_fdata_mod
       use sqrm_sdata_mod
       type(sqrm_front_type) :: front
       type(sqrm_rhs_type)   :: front_rhs
       character(len=*)      :: transp
       integer               :: br, bc
       real(r32)          :: lhs(:,:)
       integer               :: info
     end subroutine sqrm_spfct_unmqr_clean_block
  end interface qrm_spfct_unmqr_clean_block

  interface qrm_spfct_trsm_assemble_front
     subroutine sqrm_spfct_trsm_assemble_front(qrm_dscr, front, front_rhs, parent, parent_rhs, transp)
       use sqrm_fdata_mod
       use sqrm_sdata_mod
       type(qrm_dscr_type)   :: qrm_dscr
       type(sqrm_front_type) :: front
       type(sqrm_rhs_type)   :: front_rhs
       type(sqrm_front_type) :: parent
       type(sqrm_rhs_type)   :: parent_rhs
       character(len=*)      :: transp
     end subroutine sqrm_spfct_trsm_assemble_front
  end interface qrm_spfct_trsm_assemble_front
  
  interface qrm_spfct_unmqr_assemble_front
     subroutine sqrm_spfct_unmqr_assemble_front(qrm_dscr, front, front_rhs, parent, parent_rhs, transp)
       use sqrm_fdata_mod
       use sqrm_sdata_mod
       type(qrm_dscr_type)   :: qrm_dscr
       type(sqrm_front_type) :: front
       type(sqrm_rhs_type)   :: front_rhs
       type(sqrm_front_type) :: parent
       type(sqrm_rhs_type)   :: parent_rhs
       character(len=*)      :: transp
     end subroutine sqrm_spfct_unmqr_assemble_front
  end interface qrm_spfct_unmqr_assemble_front
  
end module sqrm_solve_mod
