!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine cqrm_spfct_trsm_activate_front(qrm_spfct, front, front_slv, nb, info)
  use cqrm_dsmat_mod
  use cqrm_spfct_mod
  use cqrm_sdata_mod
  use qrm_parameters_mod
  use qrm_common_mod
  use cqrm_solve_mod, protect => cqrm_spfct_trsm_activate_front
  implicit none

  type(cqrm_spfct_type) :: qrm_spfct
  type(cqrm_front_type) :: front
  type(cqrm_rhs_type)   :: front_slv
  integer               :: nb
  integer               :: info

  logical               :: seq
  
  integer                     :: err
  character(len=*), parameter :: name='qrm_spfct_trsm_activate_front'

  err = 0
  if(front%ne .le. 0) then
     front_slv%nr = 0 
     front_slv%nc = 0 
     return
  end if

  ! call cqrm_dsmat_print(front%f, 'f12.8', zero=.true., m=front%npiv, l=front%npiv)
  ! write(*,'(" ")')
  
  if(min(front%m,front%n).le.0) return
  if(front_slv%rhs%inited) return
  
  front_slv%small = front%small
  seq = front%small.ne.0

  call qrm_build_fc(front_slv%rhs%f,         &
       max(front%m, front%n), nb,            &
       front_slv%rhs%nbr, front_slv%rhs%nbc, &
       mb=front%mb,                          &
       npiv=front%sp)
  
  call cqrm_dsmat_init(front_slv%rhs, &
       max(front%m, front%n), nb,        &
       front%mb,                         &
       info=err, seq=seq)
  __QRM_INFO_CHECK(err, name,'qrm_dsmat_init',9999)

  front_slv%nr = front_slv%rhs%nbr
  front_slv%nc = front_slv%rhs%nbc


9999 continue

  info = err
  return
  
end subroutine cqrm_spfct_trsm_activate_front
