module cqrm_spfct_unmqr_task_mod

  interface qrm_spfct_unmqr_subtree_task
     module procedure cqrm_spfct_unmqr_subtree_task
  end interface qrm_spfct_unmqr_subtree_task

  interface qrm_spfct_unmqr_init_front_task
     module procedure cqrm_spfct_unmqr_init_front_task
  end interface qrm_spfct_unmqr_init_front_task

  interface qrm_spfct_unmqr_clean_front_task
     module procedure cqrm_spfct_unmqr_clean_front_task
  end interface qrm_spfct_unmqr_clean_front_task

  interface qrm_spfct_unmqr_init_block_task
     module procedure cqrm_spfct_unmqr_init_block_task
  end interface qrm_spfct_unmqr_init_block_task

  interface qrm_spfct_unmqr_clean_block_task
     module procedure cqrm_spfct_unmqr_clean_block_task
  end interface qrm_spfct_unmqr_clean_block_task

contains

  subroutine cqrm_spfct_unmqr_subtree_task(qrm_dscr, qrm_spfct, root, sdata, transp, prio)
#if defined (have_starpu)
    use fstarpu_mod
    use cqrm_starpu_mod
#endif
    use cqrm_spfct_mod
    use cqrm_fdata_mod
    use cqrm_sdata_mod
    use iso_c_binding
    use qrm_dscr_mod
    use cqrm_solve_mod
    use qrm_string_mod
    implicit none

    type(qrm_dscr_type), target         :: qrm_dscr
    type(cqrm_spfct_type), target       :: qrm_spfct
    integer, target                     :: root
    type(cqrm_sdata_type), target       :: sdata
    character(len=*), target            :: transp
    integer                             :: prio


    type(cqrm_front_type), pointer      :: front
    type(cqrm_rhs_type), pointer        :: front_slv, parent_rhs
    integer, target                     :: nh
    type(c_ptr), target                 :: qrm_spfct_c, dscrpt, sd_ptr
    integer, target                     :: dosubprio
    type(c_ptr)                         :: hdls

    ! error management
    integer                             :: info
    character(len=*), parameter         :: name='qrm_spfct_unmqr_subtree_task'

    if(qrm_dscr%info.ne.0) return
    info = 0

    front      => qrm_spfct%fdata%front_list(root)
    front_slv  => sdata%front_slv(root)
    parent_rhs => sdata%front_slv(qrm_spfct%adata%parent(root))

#if defined(have_starpu)

    if(qrm_dscr%ngpu>0) then
       dosubprio = 0
    else
       dosubprio = prio
    end if

    qrm_spfct_c = c_loc(qrm_spfct)
    dscrpt      = c_loc(qrm_dscr)
    sd_ptr      = c_loc(sdata)

    if(qrm_str_tolower(transp(1:1)) .ne. qrm_conj_transp) then
       nh = 4
    else
       nh = 3
    end if

    hdls = fstarpu_data_descr_array_alloc(nh)

    call fstarpu_data_descr_array_set(hdls, 0, &
         sdata%work%hdl, FSTARPU_SCRATCH)
    call fstarpu_data_descr_array_set(hdls, 1, &
         front%sym_handle, FSTARPU_R)
    call fstarpu_data_descr_array_set(hdls, 2, &
         front_slv%hdl, FSTARPU_RW)

    if(qrm_str_tolower(transp(1:1)) .ne. qrm_conj_transp) then
       ! top-down: add parent handle in R mode
       ! write(*,*)'   --- adding parent handle', front%num, qrm_spfct%adata%parent(root), &
            ! parent_rhs%hdl
       call fstarpu_data_descr_array_set(hdls, 3, &
            parent_rhs%hdl, FSTARPU_R)
    end if

    call fstarpu_insert_task((/ cqrm_spfct_unmqr_subtree_cl,           &
         FSTARPU_VALUE    , c_loc(dscrpt)      , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(qrm_spfct_c) , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(sd_ptr)      , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(root)        , FSTARPU_SZ_C_INT,     &
         FSTARPU_VALUE    , c_loc(transp)      , FSTARPU_SZ_CHARACTER, &
         FSTARPU_DATA_MODE_ARRAY, hdls, c_loc(nh),                     &
         FSTARPU_PRIORITY , c_loc(dosubprio),                          &
         FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                       &
         C_NULL_PTR /))

    call fstarpu_data_descr_array_free(hdls)

#else
    call cqrm_spfct_unmqr_subtree(qrm_spfct, root, sdata, transp, sdata%work, info)
    if(info.ne.0) qrm_dscr%info = info
#endif

    return

  end subroutine cqrm_spfct_unmqr_subtree_task

#if defined (have_starpu)
  recursive subroutine cqrm_spfct_unmqr_subtree_cpu_func(buffers, cl_arg) bind(C)
    use qrm_common_mod
    use cqrm_spfct_mod
    use cqrm_sdata_mod
    use cqrm_solve_mod
    use fstarpu_mod
    use iso_c_binding
    use qrm_dscr_mod
    implicit none

    type(c_ptr), value             :: cl_arg
    type(c_ptr), value             :: buffers

    type(c_ptr), target            :: qrm_spfct_c, qrm_dscr_c
    type(c_ptr), target            :: sd_ptr
    integer(c_int), target         :: root
    real(r32), target              :: flops
    character, target              :: transp

    type(qrm_dscr_type), pointer   :: qrm_dscr
    type(cqrm_spfct_type), pointer :: qrm_spfct
    type(cqrm_sdata_type), pointer :: sdata
    integer                        :: info, nw, ldw
    type(cqrm_ws_type)             :: work


    sd_ptr = c_null_ptr

    call fstarpu_unpack_arg(cl_arg,(/ c_loc(qrm_dscr_c), &
         c_loc(qrm_spfct_c),                             &
         c_loc(sd_ptr),                                  &
         c_loc(root),                                    &
         c_loc(transp) /))

    nw   = fstarpu_matrix_get_ny(buffers, 0)
    ldw  = fstarpu_matrix_get_ld(buffers, 0)
    call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), work%c, shape=(/ldw,nw/))

    call c_f_pointer(qrm_dscr_c, qrm_dscr)

    ! until StarPU has a proper error handling method we have to check
    ! manually
    if(qrm_dscr%info.ne.0) return

    call c_f_pointer(qrm_spfct_c, qrm_spfct)
    call c_f_pointer(sd_ptr, sdata)

    call cqrm_spfct_unmqr_subtree(qrm_spfct, root, sdata, transp, work, info)
    if(info.ne.0) qrm_dscr%info=info

#if defined(prnt_tsk)
    write(*,'("trsm subt  --  inode: ",i4)')root
#endif

    return

  end subroutine cqrm_spfct_unmqr_subtree_cpu_func
#endif


  subroutine cqrm_spfct_unmqr_init_front_task(qrm_dscr, qrm_spfct, sdata, fnum, transp, prio)
#if defined (have_starpu)
    use fstarpu_mod
    use cqrm_starpu_mod
#endif
    use cqrm_spfct_mod
    use cqrm_fdata_mod
    use cqrm_sdata_mod
    use iso_c_binding
    use qrm_dscr_mod
    use cqrm_solve_mod
    use qrm_string_mod
    implicit none

    type(qrm_dscr_type), target         :: qrm_dscr
    type(cqrm_spfct_type), target       :: qrm_spfct
    integer, target                     :: fnum
    type(cqrm_sdata_type), target       :: sdata
    character(len=*)                    :: transp
    integer                             :: prio

    type(cqrm_front_type), pointer      :: front
    type(cqrm_rhs_type), pointer        :: front_slv, parent_rhs
    integer, target                     :: nh, c, i, nc
    type(c_ptr), target                 :: qrm_spfct_c, dscrpt, sd_ptr
    integer, target                     :: initfrontprio
    type(c_ptr)                         :: hdls

    ! error management
    integer                             :: info
    character(len=*), parameter         :: name='qrm_spfct_unmqr_init_front_task'

    if(qrm_dscr%info.ne.0) return
    info = 0

    front      => qrm_spfct%fdata%front_list(fnum)
    front_slv  => sdata%front_slv(fnum)

#if defined(have_starpu)

    if(qrm_dscr%ngpu>0) then
       initfrontprio = 0
    else
       initfrontprio = prio
    end if

    qrm_spfct_c = c_loc(qrm_spfct)
    dscrpt      = c_loc(qrm_dscr)
    sd_ptr      = c_loc(sdata)

    nh = 2
    if((qrm_str_tolower(transp(1:1)).ne.qrm_conj_transp) .and. &
         (qrm_spfct%adata%parent(fnum).ne.0)) then
       nh = nh+1
    else
       nc = qrm_spfct%adata%childptr(fnum+1)-qrm_spfct%adata%childptr(fnum)
       nh = nh+nc
    end if

    hdls = fstarpu_data_descr_array_alloc(nh)

    call fstarpu_data_descr_array_set(hdls, 0, &
         front%sym_handle, FSTARPU_R)
    call fstarpu_data_descr_array_set(hdls, 1, &
         front_slv%hdl, FSTARPU_RW)

    if((qrm_str_tolower(transp(1:1)).ne.qrm_conj_transp) .and. &
         (qrm_spfct%adata%parent(fnum).ne.0)) then
       ! top-down: add parent handle in R mode
       parent_rhs => sdata%front_slv(qrm_spfct%adata%parent(fnum))
       call fstarpu_data_descr_array_set(hdls, 2, &
            parent_rhs%hdl, FSTARPU_R)
    else
       do i=1, nc
          c = qrm_spfct%adata%child(i-1 + qrm_spfct%adata%childptr(fnum))
          call fstarpu_data_descr_array_set(hdls, i+1, &
               sdata%front_slv(c)%hdl, FSTARPU_R)
       end do
    end if

    call fstarpu_insert_task((/ cqrm_spfct_unmqr_init_front_cl,               &
         FSTARPU_VALUE    , c_loc(dscrpt)      , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(qrm_spfct_c) , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(sd_ptr)      , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(fnum)        , FSTARPU_SZ_C_INT,     &
         FSTARPU_DATA_MODE_ARRAY, hdls, c_loc(nh),                     &
         FSTARPU_PRIORITY , c_loc(initfrontprio),                      &
         FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                       &
         C_NULL_PTR /))

    call fstarpu_data_descr_array_free(hdls)

#else
    call cqrm_spfct_unmqr_init_front(qrm_spfct, front, front_slv, info)
    if(info.ne.0) qrm_dscr%info = info
#endif

    return

  end subroutine cqrm_spfct_unmqr_init_front_task

#if defined (have_starpu)
  recursive subroutine cqrm_spfct_unmqr_init_front_cpu_func(buffers, cl_arg) bind(C)
    use qrm_common_mod
    use cqrm_spfct_mod
    use cqrm_sdata_mod
    use cqrm_solve_mod
    use fstarpu_mod
    use iso_c_binding
    use qrm_dscr_mod
    implicit none

    type(c_ptr), value             :: cl_arg
    type(c_ptr), value             :: buffers

    type(c_ptr), target            :: qrm_spfct_c, qrm_dscr_c
    type(c_ptr), target            :: sd_ptr
    integer(c_int), target         :: fnum

    type(qrm_dscr_type), pointer   :: qrm_dscr
    type(cqrm_spfct_type), pointer :: qrm_spfct
    type(cqrm_sdata_type), pointer :: sdata
    integer                        :: info
    type(cqrm_front_type), pointer :: front
    type(cqrm_rhs_type), pointer   :: front_slv

    sd_ptr = c_null_ptr

    call fstarpu_unpack_arg(cl_arg,(/ c_loc(qrm_dscr_c), &
         c_loc(qrm_spfct_c),                             &
         c_loc(sd_ptr),                                  &
         c_loc(fnum) /))

    call c_f_pointer(qrm_dscr_c, qrm_dscr)
    ! until StarPU has a proper error handling method we have to check
    ! manually
    if(qrm_dscr%info.ne.0) return

    call c_f_pointer(qrm_spfct_c, qrm_spfct)
    call c_f_pointer(sd_ptr, sdata)

    front      => qrm_spfct%fdata%front_list(fnum)
    front_slv  => sdata%front_slv(fnum)

    call cqrm_spfct_unmqr_init_front(qrm_spfct, front, front_slv, info)
    if(info.ne.0) qrm_dscr%info=info

#if defined(prnt_tsk)
    write(*,'("trsm init front  --  inode: ",i4)')fnum
#endif

    return

  end subroutine cqrm_spfct_unmqr_init_front_cpu_func
#endif



  subroutine cqrm_spfct_unmqr_init_block_task(qrm_dscr, front, front_slv, sdata, transp, br, bc, prio)
    use iso_c_binding
    use cqrm_spfct_mod
    use cqrm_fdata_mod
    use cqrm_sdata_mod
    use qrm_common_mod
    use qrm_dscr_mod
    use cqrm_solve_mod
#if defined(have_starpu)
    use fstarpu_mod
    use cqrm_starpu_mod
#endif
    implicit none

    type(qrm_dscr_type), target    :: qrm_dscr
    type(cqrm_front_type), target  :: front
    type(cqrm_rhs_type), target    :: front_slv
    type(cqrm_sdata_type), target  :: sdata
    character(len=*), target       :: transp
    integer, target                :: br, bc
    integer                        :: prio

    integer                        :: info
#if defined (have_starpu)
    type(c_ptr), target            :: frt, frtrhs, dscrpt, sd_ptr
    type(c_ptr)                    :: hdls
    integer, target                :: initbprio, nh
#endif

    if(qrm_dscr%info.ne.0) return

#if defined (have_starpu)

    frt         = c_loc(front)
    frtrhs      = c_loc(front_slv)
    dscrpt      = c_loc(qrm_dscr)
    sd_ptr      = c_loc(sdata)

    if(qrm_dscr%ngpu>0) then
       initbprio = 0
    else
       initbprio = prio
    end if

    if(front%small.eq.0) then
       nh = 2
       hdls = fstarpu_data_descr_array_alloc(nh)
       ! add front handle
       call fstarpu_data_descr_array_set(hdls, 0, &
            front_slv%hdl, FSTARPU_R)

       ! add block handle
       call fstarpu_data_descr_array_set(hdls, 1, &
            front_slv%rhs%blocks(br,bc)%hdl, FSTARPU_RW)
    else
       nh = 1
       hdls = fstarpu_data_descr_array_alloc(nh)
       ! add front handle
       call fstarpu_data_descr_array_set(hdls, 0, &
            front_slv%hdl, FSTARPU_RW)
    end if


    call fstarpu_insert_task((/ cqrm_spfct_unmqr_init_block_cl,        &
         FSTARPU_VALUE    , c_loc(dscrpt)      , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(frt)         , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(frtrhs)      , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(sd_ptr)      , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(br)          , FSTARPU_SZ_C_INT,     &
         FSTARPU_VALUE    , c_loc(bc)          , FSTARPU_SZ_C_INT,     &
         FSTARPU_VALUE    , c_loc(transp)      , FSTARPU_SZ_CHARACTER, &
         FSTARPU_DATA_MODE_ARRAY, hdls, c_loc(nh),                     &
         FSTARPU_PRIORITY , c_loc(initbprio),                          &
         FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                       &
         C_NULL_PTR /))

    call fstarpu_data_descr_array_free(hdls)
    
#else
    call cqrm_spfct_unmqr_init_block(front, front_slv, transp, br, bc, sdata%rhs, qrm_dscr%info)
#endif

    return

  end subroutine cqrm_spfct_unmqr_init_block_task


#if defined (have_starpu)
  recursive subroutine cqrm_spfct_unmqr_init_block_cpu_func(buffers, cl_arg) bind(C)
    use qrm_common_mod
    use cqrm_solve_mod
    use fstarpu_mod
    use iso_c_binding
    use qrm_dscr_mod
    use cqrm_sdata_mod
    implicit none

    type(c_ptr), value             :: cl_arg
    type(c_ptr), value             :: buffers

    type(c_ptr), target            :: frt, frtrhs, qrm_dscr_c, sd_ptr
    integer(c_int), target         :: br, bc
    character, target              :: transp

    type(cqrm_front_type), pointer :: front
    type(cqrm_rhs_type), pointer   :: front_slv
    type(qrm_dscr_type), pointer   :: qrm_dscr
    type(cqrm_sdata_type), pointer :: sdata
    integer                        :: info

    call fstarpu_unpack_arg(cl_arg,(/ c_loc(qrm_dscr_c), &
         c_loc(frt),                                     &
         c_loc(frtrhs),                                  &
         c_loc(sd_ptr),                                  &
         c_loc(br), c_loc(bc),                           &
         c_loc(transp) /))

    ! until StarPU has a proper error handling method we have to check
    ! manually
    call c_f_pointer(qrm_dscr_c, qrm_dscr)
    if(qrm_dscr%info.ne.0) return

    call c_f_pointer(frt   , front)
    call c_f_pointer(frtrhs, front_slv)
    call c_f_pointer(sd_ptr, sdata)

    call cqrm_spfct_unmqr_init_block(front, front_slv, transp, br, bc, sdata%rhs, qrm_dscr%info)

#if defined(prnt_tsk)
    write(*,'("Init trsm block  --  inode: ",i4,"  (",i4,",",i4,")")')front%num,br,bc
#endif

    return

  end subroutine cqrm_spfct_unmqr_init_block_cpu_func
#endif



  subroutine cqrm_spfct_unmqr_clean_block_task(qrm_dscr, front, front_slv, sdata, transp, br, bc, prio)
    use iso_c_binding
    use cqrm_spfct_mod
    use cqrm_fdata_mod
    use cqrm_sdata_mod
    use qrm_common_mod
    use qrm_dscr_mod
    use cqrm_solve_mod
#if defined(have_starpu)
    use fstarpu_mod
    use cqrm_starpu_mod
#endif
    implicit none

    type(qrm_dscr_type), target    :: qrm_dscr
    type(cqrm_front_type), target  :: front
    type(cqrm_rhs_type), target    :: front_slv
    type(cqrm_sdata_type), target  :: sdata
    character(len=*), target       :: transp
    integer, target                :: br, bc
    integer                        :: prio

    integer                        :: info
#if defined (have_starpu)
    type(c_ptr), target            :: frt, frtrhs, dscrpt, sd_ptr
    type(c_ptr)                    :: hdls
    integer, target                :: initbprio, nh
#endif

    if(qrm_dscr%info.ne.0) return

#if defined (have_starpu)

    frt         = c_loc(front)
    frtrhs      = c_loc(front_slv)
    dscrpt      = c_loc(qrm_dscr)
    sd_ptr      = c_loc(sdata)

    if(qrm_dscr%ngpu>0) then
       initbprio = 0
    else
       initbprio = prio
    end if

    if(front%small.eq.0) then
       nh = 2
       hdls = fstarpu_data_descr_array_alloc(nh)
       ! add front handle
       call fstarpu_data_descr_array_set(hdls, 0, &
            front_slv%hdl, FSTARPU_R)

       ! add block handle
       call fstarpu_data_descr_array_set(hdls, 1, &
            front_slv%rhs%blocks(br,bc)%hdl, FSTARPU_RW)
    else
       nh = 1
       hdls = fstarpu_data_descr_array_alloc(nh)
       ! add front handle
       call fstarpu_data_descr_array_set(hdls, 0, &
            front_slv%hdl, FSTARPU_RW)
    end if

    call fstarpu_insert_task((/ cqrm_spfct_unmqr_clean_block_cl,       &
         FSTARPU_VALUE    , c_loc(dscrpt)      , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(frt)         , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(frtrhs)      , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(sd_ptr)      , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(br)          , FSTARPU_SZ_C_INT,     &
         FSTARPU_VALUE    , c_loc(bc)          , FSTARPU_SZ_C_INT,     &
         FSTARPU_VALUE    , c_loc(transp)      , FSTARPU_SZ_CHARACTER, &
         FSTARPU_DATA_MODE_ARRAY, hdls, c_loc(nh),                     &
         FSTARPU_PRIORITY , c_loc(initbprio),                          &
         FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                       &
         C_NULL_PTR /))

    call fstarpu_data_descr_array_free(hdls)
    
#else
    call cqrm_spfct_unmqr_clean_block(front, front_slv, transp, br, bc, sdata%lhs, qrm_dscr%info)
#endif

    return

  end subroutine cqrm_spfct_unmqr_clean_block_task


#if defined (have_starpu)
  recursive subroutine cqrm_spfct_unmqr_clean_block_cpu_func(buffers, cl_arg) bind(C)
    use qrm_common_mod
    use cqrm_solve_mod
    use fstarpu_mod
    use iso_c_binding
    use qrm_dscr_mod
    use cqrm_sdata_mod
    implicit none

    type(c_ptr), value             :: cl_arg
    type(c_ptr), value             :: buffers

    type(c_ptr), target            :: frt, frtrhs, qrm_dscr_c, sd_ptr
    integer(c_int), target         :: br, bc
    character, target              :: transp

    type(cqrm_front_type), pointer :: front
    type(cqrm_rhs_type), pointer   :: front_slv
    type(qrm_dscr_type), pointer   :: qrm_dscr
    type(cqrm_sdata_type), pointer :: sdata
    integer                        :: info

    call fstarpu_unpack_arg(cl_arg,(/ c_loc(qrm_dscr_c), &
         c_loc(frt),                                     &
         c_loc(frtrhs),                                  &
         c_loc(sd_ptr),                                  &
         c_loc(br), c_loc(bc),                           &
         c_loc(transp) /))

    ! until StarPU has a proper error handling method we have to check
    ! manually
    call c_f_pointer(qrm_dscr_c, qrm_dscr)
    if(qrm_dscr%info.ne.0) return

    call c_f_pointer(frt   , front)
    call c_f_pointer(frtrhs, front_slv)
    call c_f_pointer(sd_ptr, sdata)

    call cqrm_spfct_unmqr_clean_block(front, front_slv, transp, br, bc, sdata%lhs, qrm_dscr%info)

#if defined(prnt_tsk)
    write(*,'("Clean trsm block  --  inode: ",i4,"  (",i4,",",i4,")")')front%num,br,bc
#endif

    return

  end subroutine cqrm_spfct_unmqr_clean_block_cpu_func
#endif



  subroutine cqrm_spfct_unmqr_clean_front_task(qrm_dscr, front, front_slv, sdata, transp, prio)
#if defined (have_starpu)
    use fstarpu_mod
    use cqrm_starpu_mod
#endif
    use cqrm_spfct_mod
    use cqrm_fdata_mod
    use cqrm_sdata_mod
    use iso_c_binding
    use qrm_dscr_mod
    use cqrm_solve_mod
    use qrm_string_mod
    implicit none

    type(qrm_dscr_type), target         :: qrm_dscr
    type(cqrm_front_type), target       :: front
    type(cqrm_rhs_type), target         :: front_slv
    type(cqrm_sdata_type), target       :: sdata
    character(len=*), target            :: transp
    integer                             :: prio

    integer, target                     :: nh, c, i, nc
    type(c_ptr), target                 :: frt, frtrhs, dscrpt, sd_ptr
    integer, target                     :: cleanfrontprio
    type(c_ptr)                         :: hdls

    ! error management
    integer                             :: info
    character(len=*), parameter         :: name='qrm_spfct_unmqr_clean_front_task'

    if(qrm_dscr%info.ne.0) return
    info = 0

#if defined(have_starpu)

    if(qrm_dscr%ngpu>0) then
       cleanfrontprio = 0
    else
       cleanfrontprio = prio
    end if

    dscrpt      = c_loc(qrm_dscr)
    sd_ptr      = c_loc(sdata)
    frt         = c_loc(front)
    frtrhs      = c_loc(front_slv)
    ! write(*,*)'  --- clean front task ', front%num, front_slv%hdl
    call fstarpu_insert_task((/ cqrm_spfct_unmqr_clean_front_cl,               &
         FSTARPU_VALUE    , c_loc(dscrpt)       , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(frt)          , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(frtrhs)       , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(sd_ptr)       , FSTARPU_SZ_C_PTR,     &
         FSTARPU_VALUE    , c_loc(transp)       , FSTARPU_SZ_CHARACTER, &
         FSTARPU_RW       , front_slv%hdl       ,                       &
         FSTARPU_PRIORITY , c_loc(cleanfrontprio),                      &
         FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx) ,                       &
         C_NULL_PTR /))

#else
    call cqrm_spfct_unmqr_clean_front(front, front_slv, sdata%lhs, transp, qrm_dscr%info)
#endif

    return

  end subroutine cqrm_spfct_unmqr_clean_front_task

#if defined (have_starpu)
  recursive subroutine cqrm_spfct_unmqr_clean_front_cpu_func(buffers, cl_arg) bind(C)
    use qrm_common_mod
    use cqrm_spfct_mod
    use cqrm_sdata_mod
    use cqrm_solve_mod
    use fstarpu_mod
    use iso_c_binding
    use qrm_dscr_mod
    implicit none

    type(c_ptr), value             :: cl_arg
    type(c_ptr), value             :: buffers

    type(c_ptr), target            :: qrm_dscr_c, frt, frtrhs, sd_ptr

    type(qrm_dscr_type), pointer   :: qrm_dscr
    type(cqrm_sdata_type), pointer :: sdata
    integer                        :: info
    type(cqrm_front_type), pointer :: front
    type(cqrm_rhs_type), pointer   :: front_slv
    character, target              :: transp

    call fstarpu_unpack_arg(cl_arg,(/ c_loc(qrm_dscr_c), &
         c_loc(frt),                                     &
         c_loc(frtrhs),                                  &
         c_loc(sd_ptr),                                  &
         c_loc(transp) /))

    call c_f_pointer(qrm_dscr_c, qrm_dscr)
    ! until StarPU has a proper error handling method we have to check
    ! manually
    if(qrm_dscr%info.ne.0) return

    call c_f_pointer(frt   , front)
    call c_f_pointer(frtrhs, front_slv)
    call c_f_pointer(sd_ptr, sdata)

    call cqrm_spfct_unmqr_clean_front(front, front_slv, sdata%lhs, transp, info)
    if(info.ne.0) qrm_dscr%info=info

#if defined(prnt_tsk)
    write(*,'("trsm clean front  --  inode: ",i4)')fnum
#endif

    return

  end subroutine cqrm_spfct_unmqr_clean_front_cpu_func
#endif



end module cqrm_spfct_unmqr_task_mod
