module qrm_log_mod

  use qrm_parameters_mod
  
  use iso_fortran_env

  
  type qrm_log_type
     character(len=:), allocatable :: lab, slab, plab, pslab
     character(len=:), allocatable :: val, pval
     integer                       :: verb
     logical                       :: inited=.false.
     logical                       :: sep=.false.
  end type qrm_log_type

  interface qrm_log_record
     module procedure qrm_log_drecord, qrm_log_irecord, qrm_log_srecord, qrm_log_arecord
     module procedure qrm_log_i8record, qrm_log_record_sep
  end interface qrm_log_record
  
  type(qrm_log_type), allocatable :: qrm_log_entries(:)
  integer :: nentries, largest_lab
  integer, parameter :: reallocsize=1, seplength=70, ll=20

contains


  subroutine qrm_log_init()
    implicit none

    nentries = 0
    largest_lab = 0

    if(.not.allocated(qrm_log_entries)) allocate(qrm_log_entries(reallocsize))

  end subroutine qrm_log_init


  subroutine qrm_log_drecord(lab, slab, fmt, val, verb)
    implicit none
    character*(*)     :: lab, slab, fmt
    real(real64)      :: val
    integer, optional :: verb

    integer :: e, i, l, pad
    character(len=100) :: str,fmt2
    
    
    write(str,'('//fmt//')') val
    write(fmt2,'("a",i0)')len(trim(str))
    call qrm_log_arecord(lab, slab, fmt2, trim(str), verb)
    return
  end subroutine qrm_log_drecord

  subroutine qrm_log_srecord(lab, slab, fmt, val, verb)
    implicit none
    character*(*)    :: lab, slab, fmt
    real(real32)     :: val
    integer, optional :: verb

    integer :: e, i, l, pad
    character(len=100) :: str,fmt2
    
    
    write(str,'('//fmt//')') val
    write(fmt2,'("a",i0)')len(trim(str))
    call qrm_log_arecord(lab, slab, fmt2, trim(str), verb)
    return
  end subroutine qrm_log_srecord

  subroutine qrm_log_irecord(lab, slab, fmt, val, verb)
    implicit none
    character*(*)    :: lab, slab, fmt
    integer          :: val
    integer, optional :: verb

    integer :: e, i, l, pad
    character(len=100) :: str,fmt2
    
    
    write(str,'('//fmt//')') val
    write(fmt2,'("a",i0)')len(trim(str))
    call qrm_log_arecord(lab, slab, fmt2, trim(str), verb)
    return
  end subroutine qrm_log_irecord

  subroutine qrm_log_i8record(lab, slab, fmt, val, verb)
    implicit none
    character*(*)    :: lab, slab, fmt
    integer(8)       :: val
    integer, optional :: verb

    integer :: e, i, l, pad
    character(len=100) :: str,fmt2
    
    
    write(str,'('//fmt//')') val
    write(fmt2,'("a",i0)')len(trim(str))
    call qrm_log_arecord(lab, slab, fmt2, trim(str), verb)
    return
  end subroutine qrm_log_i8record

  

  subroutine qrm_log_arecord(lab, slab, fmt, val, verb)
    implicit none
    character*(*)    :: lab, slab, fmt, val
    integer, optional :: verb

    integer :: e, i, l, pad, iverb
    type(qrm_log_type), allocatable :: tmp(:)
    character(len=100) :: str

    if(present(verb)) then
       iverb = verb
    else
       iverb = 0
    end if
    
    do e=1, nentries
       if( qrm_log_entries(e)%lab.eq.lab .and. &
            qrm_log_entries(e)%slab.eq.slab ) then
          exit
       end if
    end do

    if(e.gt.nentries) then
       ! new entry, check if there is space
       if(size(qrm_log_entries) .le. nentries) then
          ! reallocate qrm_log_entries
          allocate(tmp(nentries+reallocsize))
          tmp(1:nentries) = qrm_log_entries(1:nentries)
          deallocate(qrm_log_entries)
          call move_alloc(from=tmp, to=qrm_log_entries)
       end if
       nentries = nentries+1
    end if
    
    qrm_log_entries(e)%lab    = lab
    qrm_log_entries(e)%slab   = slab
    qrm_log_entries(e)%inited = .true.
    qrm_log_entries(e)%verb   = iverb

    if(len(lab).gt.largest_lab) then
       largest_lab=len(lab)
       do i=1, nentries-1
          if(.not.qrm_log_entries(i)%sep) then
             qrm_log_entries(i)%plab = qrm_log_entries(i)%plab//&
                  repeat(' ',largest_lab-len(qrm_log_entries(i)%plab))
          end if
       end do
    end if
    
    qrm_log_entries(e)%plab = lab//repeat(' ',largest_lab-len(lab))

    write(str,'('//fmt//')') val
    qrm_log_entries(e)%val = trim(str)

    if(fmt(1:2).eq.'es') then
       i = index(fmt, '.')
       if(i.eq.0) then
          i = len(fmt)+1
       end if
       read(fmt(3:i-1),*)l
    else if (fmt(1:1).eq.'e' .or. fmt(1:1).eq.'f') then
       i = index(fmt, '.')
       if(i.eq.0) then
          i = len(fmt)+1
       end if
       read(fmt(2:i-1),*)l
    else if (fmt(1:1).eq.'a') then
       i = len(fmt)+1
       read(fmt(2:i-1),*)l
    end if

    
    pad = len(slab)-l

    if(pad.ge.0) then
       qrm_log_entries(e)%pval  = repeat(' ',pad)//qrm_log_entries(e)%val
       qrm_log_entries(e)%pslab = slab
    else
       qrm_log_entries(e)%pval  = qrm_log_entries(e)%val
       qrm_log_entries(e)%pslab = repeat(' ',-pad)//slab
    end if
    
    ! write(*,'("| ",a," |")')qrm_log_entries(e)%pslab
    ! write(*,'("| ",a," |")')qrm_log_entries(e)%pval
    
    return

  end subroutine qrm_log_arecord


  subroutine qrm_log_record_sep(lab, verb)
    implicit none
    character*(*)    :: lab
    integer, optional :: verb

    integer :: e, iverb, lr
    type(qrm_log_type), allocatable :: tmp(:)

    if(present(verb)) then
       iverb = verb
    else
       iverb = 0
    end if
    
    e = nentries+1
    
    if(e.gt.nentries) then
       ! new entry, check if there is space
       if(size(qrm_log_entries) .le. nentries) then
          ! reallocate qrm_log_entries
          allocate(tmp(nentries+reallocsize))
          tmp(1:nentries) = qrm_log_entries(1:nentries)
          deallocate(qrm_log_entries)
          call move_alloc(from=tmp, to=qrm_log_entries)
       end if
       nentries = nentries+1
    end if
    
    qrm_log_entries(e)%lab    = lab
    qrm_log_entries(e)%inited = .true.
    qrm_log_entries(e)%verb   = iverb
    qrm_log_entries(e)%sep    = .true.
    
    lr = max((seplength - (len(lab)+2)-ll),1)
    
    qrm_log_entries(e)%plab   = achar(27)//"[31m"//repeat("=",ll)//" "//lab//" "//repeat("=",lr)//achar(27)//"[0m"

    return

  end subroutine qrm_log_record_sep

  
  subroutine qrm_log_print(mode, verb, clean, hdr)
    implicit none
 
    character*(*), optional :: mode
    integer, optional       :: verb
    logical, optional       :: clean, hdr

    integer :: i, iverb, len, ierr, ounit
    character(len=:), allocatable :: imode
    character(LEN=20) :: str
    logical           :: iclean, ihdr
    
    if(present(mode)) then
       imode = mode
    else
       imode = 'list'
    end if

    if(present(verb)) then
       iverb = verb
    else
       iverb = 0
    end if

    if(present(clean)) then
       iclean = clean
    else
       iclean = .true.
    end if

    if(present(hdr)) then
       ihdr = hdr
    else
       ihdr = .true.
    end if

    call qrm_get('qrm_ounit', ounit)
    
    if(imode.eq.'list') then
       do i=1, nentries
          if(qrm_log_entries(i)%verb.le.iverb) then
             if(qrm_log_entries(i)%sep) then
                if(ounit.gt.0) write(ounit,'(a)')qrm_log_entries(i)%plab
             else
                if(ounit.gt.0) write(ounit,'(a," : ",a)')qrm_log_entries(i)%plab,qrm_log_entries(i)%val
             end if
          end if
       end do
    else if(imode.eq.'tab') then
       i=1
       call get_environment_variable(name="QRM_LOG_HDR", value=str, length=len, status=ierr)
       if(ierr .eq. 0) read(str(1:len),*)i
       ihdr = ihdr.and.(i.gt.0)
       if(ihdr) then
          do i=1, nentries
             if(.not.qrm_log_entries(i)%sep) then
                if(qrm_log_entries(i)%verb.le.iverb) then
                   if(ounit.gt.0) write(ounit,'("| ",a," ")',advance='no')qrm_log_entries(i)%pslab
                end if
             end if
          end do
          if(ounit.gt.0) write(ounit,'("|")')
       end if
       do i=1, nentries
          if(.not.qrm_log_entries(i)%sep) then
             if(qrm_log_entries(i)%verb.le.iverb) then
                if(ounit.gt.0) write(ounit,'("| ",a," ")',advance='no')qrm_log_entries(i)%pval
             end if
          end if
       end do
       if(ounit.gt.0) write(ounit,'("|")')
    end if

    if(iclean) nentries = 0
    
    return
  end subroutine qrm_log_print


  
end module qrm_log_mod
